/* Copyright (c) 2013 APDM.
*
* PlotSensors.js
*
* Plot data from one or two monitors from a given JSON file
*/

// The server host and port
var serverHost;
var serverPort;

// Minimum width of plot window
var minWidth = 800;

// Minimum value for plot height
var minPlotHeight = 84;

// Size of viewport
var viewWidth;
var viewHeight;

// Padding around plots
var leftPad = 50;
var rightPad = 20;
var topPad = 10;
var bottomPad = 10;
var pageBottomPad = 20

// The JSON data loaded from files
var monitorData;
var eventsData;

// Size of each graph
var plotWidth;
var plotHeight;

// The current monitors being plotted
var priMonitor;
var sndMonitor = null;

// Flags to determine which sensors get plotted
var plotAccel = false;
var plotGyro = false;
var plotMag = false;
var plotOrient = false;

// The list of the available sensor name
var availableSensors = new Array();

// List of all the plots currently being displayed.
var allPlots = [];

// The zoom behavior
var zoom;

// Position and x value under mouse at zoom
var mouseXValue;
var mouseX;

// Number of zooms
var zoomInCount = 0;

// Is script waiting for new data from the server?
var loadingNewData = false;

// Options selected
var selectedOptions = {
    removeMean: false,
    plotMagnitude: false,
    bufferPercent: 0.02,
    minRangeOrientation: 0.1,
    scale: 1,
    translate: [0,0]
};

// Utility to allow parsing of host and port from url
var getLocation = function(href) {
    var url = document.createElement("a");
    url.href = href;
    return url;
};

// Redraw all the plots if the window resizes
$(window).resize(generatePlots);

// Get the paths of the files to use
var filePaths = parseURLParams(window.location.href);

// Get half the size of the plot
var halfPlotWidth = function() {
    if (plotWidth === 0) return 0;
    return plotWidth / 2;
}


function windowSize(numPlots) {
    // Take off an extra 30 for scroll bars, This may not be an issue
    // in the RawDataView. Remove the height of the options area.
    viewWidth = $(window).width() - 30;
    viewHeight = $(window).height() 
                   - $("#plotting_options_container").height() - 30;
    if (viewWidth < minWidth) {
        viewWidth = minWidth;
    }
    var minHeight = (minPlotHeight + topPad + bottomPad) * numPlots + pageBottomPad;
    if (viewHeight < minHeight) {
        viewHeight = minHeight;
    }
};

// Determine the size of each plot, n is number of plots
function setPlotSize(n) {
    plotWidth = viewWidth - leftPad - rightPad;
    plotHeight = ((viewHeight - pageBottomPad) / n) - pageBottomPad;
}


// Get the data from the JSON files
if (filePaths.eventsFile) {
    d3.json(filePaths.eventsFile, function(error, data) {
        if (error) return console.warn(error);
        eventsData = data.events;

        // Go ahead and do the date conversion here for the events
        if (eventsData) {
            eventsData.forEach( function(d) {
                d.startTime = new Date(0).setUTCMilliseconds(d.startTime/1000);
                d.endTime = new Date(0).setUTCMilliseconds(d.endTime/1000);
            });
        }
    });
}


d3.json(filePaths.dataFile, function(error, data) {
    if (error) return console.warn(error);
    console.log("Loading data file: " + filePaths.dataFile);
    monitorData = data;

    // Populate the dropdowns
    setMonitorSelections(monitorData.monitorIds, false);
    setMonitorSelections(monitorData.monitorIds, true);

    // Create the key
    createPlotKey();

    // Make sure the the starting monitor matches what is selected in
    // drop down.
    priMonitor = findMonitorByCaseId(monitorData.monitorIds[0].caseId);
    
    createPlottedSensorCheckboxes(priMonitor);
    createMeanMagCheckboxes();
    createEventsKey();
    generatePlots();

    // Remove the loading dialog
    d3.select("#loading-dialog").remove();

});

// Plot the data for a given monitor
function generatePlots() {
    // Delete the existing plot element if it exists and then create
    d3.selectAll("#plot_div").remove();

    var plotSensors = setListOfAvailableSensors(priMonitor, sndMonitor);

    // Do nothing if there are no plots to view
    if (plotSensors.numPlots < 1) return;
    
    // Get window size
    windowSize(plotSensors.numPlots);
    setPlotSize(plotSensors.numPlots);

    zoom = d3.behavior.zoom().scaleExtent([1, 100]).on("zoom", zoomChanged);
    zoom.scale(selectedOptions.scale);
    zoom.translate(selectedOptions.translate);
    var plotDiv = d3.select("body")
        .append("div")
        .attr("id", "plot_div")
        .attr("width", viewWidth)
        .attr("height", viewHeight)
        .append("svg")
        .attr("class", "plot_svg")
        .attr("width", viewWidth)
        .attr("height", viewHeight);

    var plot = d3.select(".plot_svg");

    // Create the groups for the sensor plots
    plot.selectAll("g")
        .data(plotSensors.priSensorList)
        .enter()
        .append("g")
        .attr("transform",function(d,i) {return "translate("
                                         + leftPad + ","
                                         + (topPad + (plotHeight + bottomPad) * i)
                                         + ")";})
        .attr("id", function(d) { return d.name; });

    // Plot each of the sensors
    allPlots = [];
    for (var i = 0; i < plotSensors.numPlots; i++) {
        var name = "#" + plotSensors.priSensorList[i].name;
        var sensors = { priSensor: plotSensors.priSensorList[i],
                        sndSensor: null };

        if (sndMonitor != null) {
            sensors.sndSensor = plotSensors.sndSensorList[i];
        }
        // Only put x-axis label on last SensorPlot
        var labelXAxis = false;
        if (i === plotSensors.numPlots - 1) {
            labelXAxis = true;
        }
        allPlots.push(new SensorPlot(name, sensors, eventsData, plotHeight,
                                     plotWidth, labelXAxis, selectedOptions));

    }

    // Add a clear rect to catch all the pointer events
    var mouseG = plot.append("g")
        .attr("id", "pointer-mask")
        .attr("transform", "translate(" + leftPad + ",0)");
    mouseG.append("rect")
        .attr("id", "pointer-mask-rect")
        .attr("width", plotWidth)
        .attr("height", viewHeight)
        .attr("fill", "#FFFFFF")
        .attr("fill-opacity", "0")
        .attr("pointer-events", "all")
        .call(zoom);

    createResamplingDialog();
}


// Update all the plots with the new zoom level.
function zoomChanged() {
    // Waiting to load data, don't do anything
    if (loadingNewData) return;
    
    // Need these to determine the correct translate to keep mouse
    // under cursor
    mouseX = d3.event.sourceEvent.clientX - leftPad; 
    mouseXValue = SensorPlot.xScale.invert(mouseX);

    // Check stride to see if more points are available
    if ((monitorData.monitors[0].Calibrated[0].attributes.stride[0] > 1)
        && (d3.event.scale > 3)) {
        zoomIn();
        return;

    } else if (d3.event.scale < 1.5 && zoomInCount > 0) {
        zoomOut();
        return;
    }

    // Handle a pan off the plot.
    if (zoomInCount > 0) {
        if (zoom.translate()[0]  > 0) {
            console.log("Pan off left edge...");
            loadingNewData = true;
            if (pan(false, -0.25)) {
                return;
            }
            loadingNewData = false;
        }
        if ((zoom.translate()[0] / (zoom.scale() - 1))
            < -(SensorPlot.xScale.range()[1])) {
            console.log("Pan off right edge...");
            loadingNewData = true;
            if (pan(true, 0.25)) {        // Add a quarter of the range
                return;
            }
            loadingNewData = false;
        }
    }

    selectedOptions.scale = zoom.scale();
    // Make sure that we are copying the values, not a reference
    selectedOptions.translate[0] = zoom.translate()[0];
    selectedOptions.translate[1] = zoom.translate()[1];
    allPlots.forEach( function(d) { d.update(zoom.scale(), zoom.translate()); });
}


function zoomIn() {
    var indexes = findRange(true, (halfPlotWidth() / 2), false);
    var length = plotWidth + halfPlotWidth();
    console.log("Getting new zoom IN data...");
    zoomInCount = zoomInCount + 1;
    showResamplingDialog(true);
    loadingNewData = true;
    sendDataRequest(indexes, length, false, false);
    selectedOptions.scale = 2;
}


function zoomOut() {
    var indexes = findRange(false, plotWidth * (2/3), false);
    var length = plotWidth + halfPlotWidth();
    console.log("Getting new zoom OUT data...");
    zoomInCount = zoomInCount - 1;
    showResamplingDialog(true);
    loadingNewData = true;
    sendDataRequest(indexes, length, false, false);
    selectedOptions.scale = 2.5;
}


// Get data on either the left or right off the screen
function pan(isRightPan, offset) {
    var indexes = findRange(false, offset, true);
    var length = plotWidth * 2;

    // Can't update, because we are at the end of the range
    if (!indexes) {
        return false;
    }
    showResamplingDialog(true);
    sendDataRequest(indexes, length, true, isRightPan);
    selectedOptions.scale = d3.event.scale;
    return true;
}


function sendDataRequest(indexes, length, isPan, isRightPan) {
    var senList = "[\"" + availableSensors.join("\",\"") + "\"]";
    var json =
        "{\"displayWidth\": " + length
        + ",\"sensors\": " + senList
        + ",\"startRow\": " + indexes.startIdx
        + ",\"endRow\": " + indexes.endIdx
        + "}";
    // Get new data from the server
    $.post("http://" + serverHost + ":" + serverPort
           + "/static/html/",
           {
               options: json,
               dataFile: monitorData.DataFileName,
               jsonFile: monitorData.jsonFileName
           },
           function(d) { updateMonitorData(d, isPan, isRightPan); },
           "json");
}


function findRange(zoomIn, offset, isPan) {
    var totalLen = monitorData.monitors[0].Calibrated[0].attributes.dims[0] - 1;
    var selected = monitorData.monitors[0].Calibrated[0].attributes.selectedDims[0];
    var start = monitorData.monitors[0].Calibrated[0].attributes.startDims[0];
    var stride = monitorData.monitors[0].Calibrated[0].attributes.stride[0];
    var end = (selected * stride) + start;
    var range = (end - start);
    // This assumes a scale of 1
    var pointsPerPixel = range / (plotWidth * 2);
    var startIdx;
    var endIdx;

    if (isPan) {
        // Don't do anything if we are already at the left or right end
        // datarange
        if ((start <= 0 && offset < 0)
            || ((end + stride) >= totalLen && offset > 0)) {
            return false;
        }

        var shift = range * offset;
        // Now we want a new start and end index for the new sample
        startIdx = start + shift;
        endIdx = end + shift;
    } else {
        // Now we want a new start and end index for the new sample
        if (zoomIn) {
            // Find the index into data that matches
            var xZero = SensorPlot.xScale.invert(0);
            var i = 0;
            while (monitorData.monitors[0].Calibrated[0].data[i].time / 1000 < xZero) {
                i++;
            }
            var plotStartIdx = (i * stride) + start;

            startIdx = plotStartIdx - (offset * pointsPerPixel);
            endIdx =  startIdx + ((plotWidth + offset) * pointsPerPixel);
        } else {
            startIdx = start - Math.floor(offset * pointsPerPixel);
            endIdx = end + Math.ceil(offset * pointsPerPixel);
        }
    }

    if (startIdx <= 0) {
        // Shitf the end index
        endIdx = endIdx + Math.abs(startIdx);
        startIdx = 0;

        if (endIdx >= totalLen) {
            endIdx = totalLen;
        }
    }
    if (endIdx > totalLen) {
        // Shift the start index
        startIdx = startIdx - (endIdx - totalLen);
        endIdx = totalLen;
        if (startIdx < 0) {
            startIdx = 0;
        }
    }

    return {startIdx: Math.floor(startIdx), endIdx: Math.ceil(endIdx)};
}


// Update the monitor data
function updateMonitorData(newData, isPan, isRightPan) {
    console.log("Received new data");
    monitorData = newData;

    priMonitor = findMonitorByCaseId(priMonitor.caseId);
    if (sndMonitor) {
        sndMonitor = findMonitorByCaseId(sndMonitor.caseId);
    }

    // Use generate plots since it recreates the SensorPlot objects
    // with the new data.
    generatePlots();

    // We need to call update again so that we can make sure the same
    // x value is under the mouse. Also need to update the zoom
    // behavior with the adjusted translate.
    if (!isPan) {
        var trOffset = mouseX - SensorPlot.xScale(mouseXValue);
        selectedOptions.translate[0] = selectedOptions.translate[0] + trOffset;
        zoom.scale(selectedOptions.scale);
        zoom.translate(selectedOptions.translate);
        allPlots.forEach( function(d) {
            d.update(selectedOptions.scale, selectedOptions.translate); });
    } else {
        if (isRightPan) {
            selectedOptions.translate[0] = zoom.translate()[0]
                + ((plotWidth / 4) * selectedOptions.scale);
        } else {
            selectedOptions.translate[0] = -(plotWidth / 4) * selectedOptions.scale;
        }
        zoom.translate(selectedOptions.translate);
        zoom.scale(selectedOptions.scale);
        allPlots.forEach( function(d) {
            d.update(selectedOptions.scale, selectedOptions.translate); });
    }
    loadingNewData = false;
    showResamplingDialog(false);
}


function createResamplingDialog() {
    var g = d3.select("#pointer-mask");
    g.append("rect")
        .attr("class", "dialog_resampling")
        .attr("x", ((plotWidth / 2) - 100))
        .attr("y", ((viewHeight / 2) - 50))
        .attr("width", 200).attr("height", 40)
        .attr("rx", 10).attr("ry", 10)
        .attr("opacity", 0);

    g.append("text")
        .attr("class", "dialog_resampling_text")
        .attr("x", ((plotWidth / 2) - 50))
        .attr("y", ((viewHeight / 2) - 25))
        .attr("opacity", 0)
        .text("Resampling...");
}

function showResamplingDialog(showDialog) {
    var opacity = 0;
    if (showDialog) {
        opacity = 0.90;
    }
    d3.select(".dialog_resampling")
        .attr("opacity", opacity);
    d3.select(".dialog_resampling_text")
        .attr("opacity", opacity);
}

// Return the "filepath" parameter passed with the URL. We can be a
// little lazy here since we know this will be the only param passed.
function parseURLParams(url) {
    // Find the server host and port
    serverHost = getLocation(url).hostname;
    serverPort = getLocation(url).port;

    var dataPath;
    var eventsPath;
    var start = url.indexOf("?") + 1;
    var end = url.length + 1;
    var param = url.slice(start, end - 1);

    if (param === url || param === "") return;

    var values = param.split("&");
    for (var i = 0; i < values.length; i++) {
        var nameValue = values[i].split("=");
        if (nameValue[0] === "dataFile") {
            dataPath = decodeURIComponent(nameValue[1]);
        } else if (nameValue[0] === "eventsFile") {
            eventsPath = decodeURIComponent(nameValue[1]);
        }
    }

    return { dataFile: dataPath, eventsFile: eventsPath };
};


// Handlers for changes in the monitor select dropdowns
var primarySelectChange = function () {
    console.log("Selecting " + this.value + " for plotting");
 
    priMonitor = findMonitorByCaseId(this.value);
    if (priMonitor) {
        generatePlots();
    }
}
var secondarySelectChange = function () {
    console.log("Selecting " + this.value + " for plotting");
 
    sndMonitor = findMonitorByCaseId(this.value);
    generatePlots();
}

// Populate the "Select Monitor" drop down with the available monitors
function setMonitorSelections(monitorIds, isSecondary) {
    var dropdownId = "#primary_monitor_select";
    var dropdownText = "Select Monitor: ";
    var monitors = monitorIds;
    var selectFunc = primarySelectChange;

    if (isSecondary) {
        // No comparison if only one monitor
        if (monitorIds.length <= 1)
            return;

        var monitors = new Array();
        monitors.push( { monitorLabel: "None",
                     caseId: "None" });
        monitorIds.forEach( function (d) {
            monitors.push(d);
        });
        dropdownId = "#secondary_monitor_select";
        dropdownText = "Compare to: ";
        selectFunc = secondarySelectChange;
    }
    d3.select(dropdownId)
        .text(dropdownText);

    d3.select(dropdownId)
        .append("select")
        .on("change", selectFunc);

    d3.select(dropdownId)
        .select("select").selectAll("option")
        .data(monitors)
        .enter()
        .append("option")
        .attr("value", function (d) { return d.caseId; })
        .text( function (d) {
            if (d.caseId === "None") {
                return d.caseId;
            } else {
                return ("#" + d.caseId.match(/[1-9][0-9]*$/)
                        + ": " + d.monitorLabel);
            }});
}

// Add checkboxes for each available sensor to set which are plotted
function createPlottedSensorCheckboxes(monitor) {

    // Get a list of all the sensors in the "Calibrated" section.
    var sensors = new Array();
    monitor.Calibrated.forEach( function(d) {
        sensors.push(d.name);
    });

    if (!monitor.attributes) {
        // Assume that all the sensors are active
        availableSensors = sensors.sort();
        plotAccel = true;
        plotGyro = true;
        plotMag = true;
        plotOrient = true;
    } else {
        if (monitor.attributes.AccelerometersEnabled == 1) {
            availableSensors.push("Accelerometers");
            plotAccel = true;
        }
        if (monitor.attributes.GyroscopesEnabled == 1) {
            availableSensors.push("Gyroscopes");
            plotGyro = true;
        }
        if (monitor.attributes.MagnetometersEnabled == 1) {
            availableSensors.push("Magnetometers");
            plotMag = true;
        }
        // Since we don't have enabled flag, see if its in the list
        sensors.forEach( function(d) {
            if (d === "Orientation") {
                availableSensors.push(d);
                plotOrient = true;
            }
        });
    }

    if (availableSensors.length > 0) {
        // Create checkboxes for each available sensor
        d3.select("#sensor_selection").selectAll("input")
            .data(availableSensors)
            .enter()
            .append("label")
            .text(function (d) { return d; })
            .append("input")
            .attr("checked", true)
            .attr("type", "checkbox")
            .attr("value", function (d) { return d; })
            .on("change", onPlottedSensorCheckBoxChange);
    }
}

function onPlottedSensorCheckBoxChange() {
    var sensorName = this.value;
    if (sensorName === "Accelerometers") {
        if (this.checked) {
            plotAccel = true;
        } else {
            plotAccel = false;
        }
    } else if (sensorName === "Gyroscopes") {
        if (this.checked) {
            plotGyro = true;
        } else {
            plotGyro = false;
        }
    } else if (sensorName === "Magnetometers") {
        if (this.checked) {
            plotMag = true;
        } else {
            plotMag = false;
        }
    } else if (sensorName == "Orientation") {
        if (this.checked) {
            plotOrient = true;
        } else {
            plotOrient = false;
        }
    }
    generatePlots();
}

function createMeanMagCheckboxes() {
    // Checkbox for magnitude
    var box = d3.select("#mean_mag_selection")
        .append("label")
        .text("Remove mean")
        .append("input")
        .attr("type", "checkbox")
        .attr("value", "removeMean")
        .on("change", onMeanCheckboxChange);
    if (selectedOptions.removeMean) {
        box.attr("checked", true);
    }

    box = d3.select("#mean_mag_selection")
        .append("label")
        .text("Magnitude")
        .append("input")
        .attr("type", "checkbox")
        .attr("value", "plotMagnitude")
        .on("change", onMagnitudeChange);
    if (selectedOptions.plotMagnitude) {
        box.attr("checked", true);
    }
}

function onMeanCheckboxChange() {
    if (this.checked) {
        selectedOptions.removeMean = true;
    } else {
        selectedOptions.removeMean = false;
    }
    generatePlots();
}


function onMagnitudeChange() {
    if (this.checked) {
        selectedOptions.plotMagnitude = true;
    } else {
        selectedOptions.plotMagnitude = false;
    }
    generatePlots();
}

// Find a sensor in a monitor's calibrated group by a given name
function findCalibratedSensorByName(monitor, name) {
    if (monitor == null) {
        return null;
    }
    var sensors = monitor.Calibrated;
    for (var i = 0; i < sensors.length; i++) {
        if (sensors[i].name === name) {
            return sensors[i];
        }
    }
}


// Find a monitor in monitorData by a given Case ID
function findMonitorByCaseId(id) {
    for (var i = 0; i < monitorData.monitors.length; i++) {
        if (monitorData.monitors[i].caseId === id) {
            return monitorData.monitors[i];
        }
    }
    return null;
}

// Create the list of sensors based on monitor attributes
function setListOfAvailableSensors(monitorOne, monitorTwo) {
    var priSensorList = [];
    var sndSensorList = [];
    var numPlots = 0;

    if (plotAccel) {
        sndSensorList[numPlots] = findCalibratedSensorByName(monitorTwo, "Accelerometers");
        priSensorList[numPlots++] = findCalibratedSensorByName(monitorOne, "Accelerometers");
    }
    if (plotGyro) {
        sndSensorList[numPlots] = findCalibratedSensorByName(monitorTwo, "Gyroscopes");
        priSensorList[numPlots++] = findCalibratedSensorByName(monitorOne, "Gyroscopes");
    }
    if (plotMag) {
        sndSensorList[numPlots] = findCalibratedSensorByName(monitorTwo, "Magnetometers");
        priSensorList[numPlots++] = findCalibratedSensorByName(monitorOne, "Magnetometers");
    }
    if (plotOrient) {
        sndSensorList[numPlots] = findCalibratedSensorByName(monitorTwo, "Orientation");
        priSensorList[numPlots++] = findCalibratedSensorByName(monitorOne, "Orientation");
    }

    return { priSensorList: priSensorList,
             sndSensorList: sndSensorList,
             numPlots: numPlots };
}


// Create the key for the plot
function createPlotKey() {

    var lineLength = 20;

    var key = d3.select(".plot_key")
        .append("svg")
        .attr("width", 100)
        .attr("height", 20)
        .append("g");
    
    key.append("line")
        .attr("class", "line_x")
        .attr("x1", 0).attr("y1", 10)
        .attr("x2", lineLength).attr("y2", 10);

    key.append("text")
        .attr("text-anchor", "middle")
        .attr("x", lineLength + 5)
        .attr("y", 13)
        .text("x");

    key.append("line")
        .attr("class", "line_y")
        .attr("x1", lineLength + 14).attr("y1", 10)
        .attr("x2", (lineLength + 7) * 2).attr("y2", 10);

    key.append("text")
        .attr("text-anchor", "middle")
        .attr("x", (lineLength + 7) * 2 + 5)
        .attr("y", 13)
        .text("y");

    key.append("line")
        .attr("class", "line_z")
        .attr("x1", (lineLength + 7) * 2 + 10).attr("y1", 10)
        .attr("x2", (lineLength + 7) * 3 + 3).attr("y2", 10);

    key.append("text")
        .attr("text-anchor", "middle")
        .attr("x", (lineLength + 7) * 3 + 8)
        .attr("y", 13)
        .text("z");

}


// Create the key for plotted events
function createEventsKey() {
    if (!eventsData) return;

    var numEvents = eventsData.length;
    var svg = d3.select(".event_key")
        .append("svg")
        .attr("height", 20)
        .attr("width", eventsData.length * 100);

    
    svg.selectAll("g")
        .data(eventsData)
        .enter()
        .append("g")
        .append("rect")
        .attr("class", function(d) { return "event_rect " + d.type; })
        .attr("x", function(d, i) { return i * 100; })
        .attr("y", 0)
        .attr("width", 20).attr("height", 20)

    svg.selectAll("g")
        .data(eventsData)
        .append("text")
        .attr("text-anchor", "start")
        .attr("x", function(d, i) { return (i * 100) + 24; })
        .attr("y", 13)
        .text(function(d) { return d.name; });
}


// Utility function to determine the time range of the sample
function getTimeRange(timeDiff) {
    var timeDiff = timeDiff / 1000;
    var seconds = Math.round(timeDiff % 60);
    timeDiff = timeDiff / Math.round(60);
    var minutes = Math.round(timeDiff % 60);
    timeDiff = timeDiff / Math.round(60);
    var hours = Math.round(timeDiff % 24);
    return { secondsElapased: seconds, minutesElapased: minutes,
             hoursElapased: hours };
}

// Object representing a single sensor plot
function SensorPlot(id, sensors, events, height, width, labelXAxis, options) {
    var self = this;
    self.priSensor = sensors.priSensor;
    self.sndSensor = sensors.sndSensor;
    self.events = events;
    self.height = height;
    self.width = width;
    self.options = options || {};
    self.options.removeMean = self.options.removeMean || false;
    self.options.plotMagnitude = self.options.plotMagnitude || false;
    self.options.bufferPercent = self.options.bufferPercent || 0;
    self.options.minRangeOrientation = self.options.minRangeOrientation || 0;
    self.options.scale = self.options.scale || 1;
    self.options.translate = self.options.translate || [0,0];
    self.labelXAxis = labelXAxis;
    self.isOrientation = false;
    
    if (self.priSensor.name === "Orientation") {
        self.isOrientation = true;
    }
    
    // Convert date format to miliseconds
    self.priSensor.data.forEach( function (d) {
        d.date = new Date(0).setUTCMilliseconds(d.time/1000);
    });
    
    if (self.sndSensor != null) {
        self.sndSensor.data.forEach( function (d) {
            d.date = new Date(0).setUTCMilliseconds(d.time/1000);
        });
    }

    // TODO: Since we reset the scale and translate when we call
    // update(), can the setting of scalse here be rethought/refactored?
    self.genXScale();
    if (self.options.plotMagnitude) {
        self.genMagScales();
    } else {
        self.genStandardScales();
    }

    // Create plot axis

	self.xAxisGrid = d3.svg.axis()
	    .scale(self.xScale)
	    .orient("bottom")
        .tickSize(height, 0, 0)
        .tickFormat("");

	self.yAxis = d3.svg.axis()
	    .scale(self.yScale)
        .ticks(5)
	    .orient("left");

    // Get the area we will be plotting in
    self.plotArea = d3.select(id);

    // Add axis and lines at tick marks
    self.xGrid = self.plotArea.append("g")
        .attr("class", "grid")
        .attr("transform", "translate(0,0)")
        .call(self.xAxisGrid);

    self.plotArea.append("g")
        .attr("class", "y axis")
        .attr("transform", "translate(0,0)")
        .call(self.yAxis);

    self.plotArea.append("g")
        .attr("class", "grid")
        .attr("transform", "translate(0,0)")
        .call(self.yAxis
              .tickSize(-width, 0, 0)
              .tickFormat(""));

	// Add the y-axis label
    self.plotArea.append("text")
        .attr("class", "y label")
        .attr("text-anchor", "middle")
        .attr("x", -plotHeight/2)
        .attr("y", -35)
        .attr("transform", "rotate(-90)")
        .text(function(d) {
            if (!self.priSensor.attributes.Units) {
                return self.priSensor.name;
            }
            return self.priSensor.name + " ("
                + self.priSensor.attributes.Units + ")";
             });

    if (labelXAxis) {    
        // Only draw x axis ticks and label on given plot
        xFormat = self.timeTickFormat();
	    self.xAxisTicks = d3.svg.axis()
	        .scale(self.xScale)
	        .orient("bottom")
            .ticks(5)
            .tickFormat(d3.time.format(xFormat.format));

        self.xTicks = self.plotArea.append("g")
            .attr("class", "x axis")
            .attr("transform", "translate(0, "
                  + height + ")")
            .call(self.xAxisTicks);

        self.plotArea.append("text")
            .attr("class", "x label")
            .attr("text-anchor", "middle")
            .attr("x", width/2)
            .attr("y", height + 30)
            .text(("Date/Time " + xFormat.label));
    } else {
        self.plotArea.append("line")
            .attr("class", "x_axis_line")
	        .attr("x1", 0).attr("y1", self.height)
            .attr("x2", self.width).attr("y2", self.height);
    }

    // Create view box to limit visible lines and a <g> container for
    // the lines
    self.viewBox = self.plotArea.append("svg")
        .attr("width", self.width)
        .attr("height", self.height)
        .attr("viewBox", "0 0 " + self.width + " " + self.height);

    self.update(self.options.scale, self.options.translate);
};


// Determine the format to use for time axis
SensorPlot.prototype.timeTickFormat = function() {
    var elapased = getTimeRange(this.priSensor.data[this.priSensor.data.length -1].date
                                - this.priSensor.data[0].date)
    var format = "%S.%L";
    var label = "Sec)";
    if (elapased.minutesElapased > 0) {
        format = "%M:" + format;
        label = "Min:" + label;
    }
    if (elapased.hoursElapased > 0) {
        format = "%H:" + format
        label = "Hrs:" + label;
    }
    return { format: format, label: ("(" + label) };
}

// Use "static" variables to hold the xScale function, since it is the
// same for all sensors, and we need access to it like a static
// method.
SensorPlot.xScale;


// Generate the x scale function. We assume that different monitors
// will have the same time range
SensorPlot.prototype.genXScale = function() {
    this.xMin = this.priSensor.data[0].date;
    this.xMax = this.priSensor.data[this.priSensor.data.length - 1].date;

    this.xScale = d3.time.scale.utc()
        .domain([this.xMin, this.xMax])
        .range([0, this.width]);
    this.staticXScale = this.xScale.copy();
    SensorPlot.xScale = this.xScale;
}


// Generate the scales for plotting, either with or without the mean
// subtraction
SensorPlot.prototype.genStandardScales = function() {

    var mean = { x:0, y:0, z:0, scalar:0 };

    if (this.options.removeMean) {
        mean = this.priSensor.mean || this.calculateMean(this.priSensor);
    }
    var yValues = this.priSensor.data.map(function(d)
                                                  { return d.x - mean.x; })
                       .concat(this.priSensor.data.map(function(d)
                                                       { return d.y -mean.y; }),
                               this.priSensor.data.map(function(d)
                                                       { return d.z - mean.z; }));
    if (this.isOrientation) {
        yValues = yValues.concat(this.priSensor.data.map(
            function(d) { return d.scalar - mean.scalar; }));
    }
	var yTempMin = d3.min(yValues);
	var yTempMax = d3.max(yValues);

    if (this.sndSensor != null) {
        mean = { x:0, y:0, z:0, scalar:0 };
        if (this.options.removeMean) {
            mean = this.sndSensor.mean || this.calculateMean(this.sndSensor);
        }
	    yValues = this.sndSensor.data.map(function(d) { return d.x - mean.x; })
                             .concat(this.sndSensor.data.map(function(d)
                                                             { return d.y - mean.y; }),
                                     this.sndSensor.data.map(function(d)
                                                             { return d.z - mean.z; }));
        if (this.isOrientation) {
            yValues = yValues.concat(this.sndSensor.data.map(
                function(d) { return d.scalar - mean.scalar; }));
        }
        yTempMin = d3.min([d3.min(yValues), yTempMin]);
        yTempMax = d3.max([d3.max(yValues), yTempMax]);
    }
    // Add a small percentage of the range back on, for some
    // buffer space on the plot
    var rangeBuf = (yTempMax - yTempMin) * this.options.bufferPercent;
    this.yMin = yTempMin - rangeBuf;
    this.yMax = yTempMax + rangeBuf;

    // Determine the y scale for this data
	this.yScale = d3.scale.linear()
	    .domain([this.yMin, this.yMax])
	    .range([this.height, 0]);
    this.staticYScale = this.yScale.copy();
};


// Generate the data and scales needed for a plot of sensor magnitude
SensorPlot.prototype.genMagScales = function () {
    this.calculateMagnitude(this.priSensor);
    if (this.sndSensor) {
        this.calculateMagnitude(this.sndSensor);
    }

	var yTempMin = d3.min(this.priSensor.data.map(function(d) { return d.mag; }));
	var yTempMax = d3.max(this.priSensor.data.map(function(d) { return d.mag; }));
    if (this.sndSensor != null) {
	    var ySndMin = d3.min(this.sndSensor.data.map(function(d) { return d.mag; }));
	    var ySndMax = d3.max(this.sndSensor.data.map(function(d) { return d.mag; }));
        yTempMin = d3.min([ySndMin, yTempMin]);
        yTempMax = d3.max([ySndMax, yTempMax]);
    }
    // Add a small percentage of the range back on, for some
    // buffer space on the plot
    var rangeBuf = (yTempMax - yTempMin) * this.options.bufferPercent;
    this.yMin = yTempMin - rangeBuf;
    this.yMax = yTempMax + rangeBuf;


    if (this.isOrientation) {
        var diff = this.options.minRangeOrientation - (this.yMax - this.yMin);
        if (diff > 0) {
            this.yMax = this.yMax + (diff / 2) + rangeBuf;
            this.yMin = this.yMin - (diff / 2) + rangeBuf;
        }
    }

    // Determine the y scale for this data
	this.yScale = d3.scale.linear()
	    .domain([this.yMin, this.yMax])
	    .range([this.height, 0]);
    this.staticYScale = this.yScale.copy();
};


// Update the plot lines and axis given a new scale and translate vector
SensorPlot.prototype.update = function(scale, translate) {

    this.addEventsToPlot();

    if (this.lineG) {
        this.lineG.remove();
    }
    this.lineG = this.viewBox.append("g")
        .attr("transform", "translate(" + 0 + ",0)")

    this.updateScales(scale, translate);
    // We are only handling x axis scaling at the moment
    if (this.options.plotMagnitude) {
        if (this.sndSensor != null) {
            this.addMagLinesToPlot(this.sndSensor, "secondary_lines");
        }
        this.addMagLinesToPlot(this.priSensor, "primary_lines");
    } else {
        if (this.sndSensor != null) {
            this.addLinesToPlot(this.sndSensor, "secondary_lines");
        }
        this.addLinesToPlot(this.priSensor, "primary_lines");
    }

    // Rescale the axises
    if (this.labelXAxis) {
        this.xAxisTicks.scale(this.xScale);
        this.xTicks.call(this.xAxisTicks);
    }
    this.xAxisGrid.scale(this.xScale);
    this.yAxis.scale(this.yScale);
    this.xGrid.call(this.xAxisGrid);

}


SensorPlot.prototype.updateScales = function(scale, tr) {
    var xCopy = this.staticXScale.copy();

    // Limit panning to range of data
    tr[0] = Math.min(tr[0], 0);
    tr[0] = Math.max(tr[0], this.width - this.xScale.range()[1] * scale);

    // Update the x scaling function
    this.xScale.domain(xCopy.range().map(function(x) {
        return (x - tr[0]) / scale;
    }).map(xCopy.invert));

    // TODO: Determine how we want to handle y axis scaling
    // this.yScale = this.staticYScale.copy();
    // var yCopy = this.yScale.copy();
    // // Update the x scaling function
    //   this.yScale.domain(yCopy.range().map(function(y) {
    //     return (y - translate[1]) / scale;
    //   }).map(yCopy.invert));
};
                                      

// Add the lines for the given plot. 
SensorPlot.prototype.addLinesToPlot = function(sensor, lineStyle) {
    var self = this;
    // Define methods for lines, removing the mean if necessary
    var mean = {x:0, y:0, z:0, scalar:0};
    if (this.options.removeMean) {
        mean = sensor.mean || this.calculateMean(sensor);
    }
    sensor.lineX = d3.svg.line()
        .x(function (d) { return self.xScale(d.date); })
        .y(function (d) { return self.yScale(d.x - mean.x); });

    sensor.lineY = d3.svg.line()
        .x(function (d) { return self.xScale(d.date); })
        .y(function (d) { return self.yScale(d.y - mean.y); });

    sensor.lineZ = d3.svg.line()
        .x(function (d) { return self.xScale(d.date); })
        .y(function (d) { return self.yScale(d.z - mean.z); });

    sensor.plotX = self.lineG.data([sensor.data])
        .append("path")
        .attr("class", "line_x " + lineStyle)
	    .attr("d", sensor.lineX);

    sensor.plotY = self.lineG.data([sensor.data])
        .append("path")
        .attr("class", "line_y " + lineStyle)
	    .attr("d", sensor.lineY);

    sensor.plotZ = self.lineG.data([sensor.data])
        .append("path")
        .attr("class", "line_z " + lineStyle)
	    .attr("d", sensor.lineZ);

    // 4-axis sensor
    if (sensor.name == "Orientation") {
        sensor.lineScalar = d3.svg.line()
            .x(function (d) { return self.xScale(d.date); })
            .y(function (d) { return self.yScale(d.scalar - mean.scalar); });
        sensor.plotScalar = self.lineG.data([sensor.data])
            .append("path")
            .attr("class", "line_scalar " + lineStyle)
	        .attr("d", sensor.lineScalar);
    }
};


// Generate and add the magintude lines to the plot
SensorPlot.prototype.addMagLinesToPlot = function(sensor, lineStyle) {
    var self = this;
    sensor.lineMag = d3.svg.line()
        .x(function (d) { return self.xScale(d.date); })
        .y(function (d) { return self.yScale(d.mag); });

    sensor.plotMagnitude = self.lineG.data([sensor.data])
        .append("path")
        .attr("class", "line_x " + lineStyle)
	    .attr("d", sensor.lineMag);
};


// Add the areas for events
SensorPlot.prototype.addEventsToPlot = function() {
    var self = this;
    if (!this.events) return;

    if (this.eventsG) {
        this.eventsG.remove();
    }
    this.eventsG = this.viewBox.append("g");

    this.eventsG.selectAll("g")
        .data(this.events)
        .enter()
        .append("rect")
        .attr("class", function(d) { return "event_rect " + d.type })
        .attr("x", function(d) { return self.xScale(d.startTime); })
        .attr("y", "0")
        .attr("width", function(d) { return self.xScale(d.endTime)
                                     - self.xScale(d.startTime); })
        .attr("height", function(d) { return self.height; })
    
};


// Calculate the mean of the sensor data
SensorPlot.prototype.calculateMean = function(sensor) {
    var xSum = 0;
    var ySum = 0;
    var zSum = 0;
    var scalarSum = 0;
    for (var i = 0; i < sensor.data.length; i++) {
        xSum = sensor.data[i].x + xSum;
        ySum = sensor.data[i].y + ySum;
        zSum = sensor.data[i].z + zSum;
        if (this.isOrientation) {
            scalarSum = sensor.data[i].scalar + scalarSum;
        }
    }
    sensor.mean = { x: xSum / sensor.data.length,
                    y: ySum / sensor.data.length,
                    z: zSum / sensor.data.length };
    if (this.isOrientation) {
        sensor.mean.scalar = scalarSum / sensor.data.length;
    }
    return sensor.mean;
};


// Calculate the magnitude for the given sensor and add it to is data
// structure.
SensorPlot.prototype.calculateMagnitude = function(sensor) {
    // Generate the mean for each axis, if needed. Magnitude is not
    // useful if we don't remove the mean first.
    var mean = { x:0, y:0, z:0, scalar:0 };
    if (this.options.removeMean) {
        mean = sensor.mean || this.calculateMean(sensor);
    } 

    if (this.isOrientation) {
        for (var i = 0; i < sensor.data.length; i++) {
            sensor.data[i].mag = Math.sqrt(Math.pow(sensor.data[i].x - mean.x, 2)
                                           + Math.pow(sensor.data[i].y - mean.y, 2)
                                           + Math.pow(sensor.data[i].z - mean.z, 2)
                                           + Math.pow(sensor.data[i].scalar - mean.scalar, 2));
        }
    } else {
        for (var i = 0; i < sensor.data.length; i++) {
            sensor.data[i].mag = Math.sqrt(Math.pow(sensor.data[i].x - mean.x, 2)
                                           + Math.pow(sensor.data[i].y - mean.y, 2)
                                           + Math.pow(sensor.data[i].z - mean.z, 2))
        }
    }
};
